from .imports import * 

def tee_output(file):
    tee = subprocess.Popen(["tee", file], stdin=subprocess.PIPE)
    os.dup2(tee.stdin.fileno(), sys.stdout.fileno()) # type: ignore
    os.dup2(tee.stdin.fileno(), sys.stderr.fileno()) # type: ignore


from .imports import *

# todo this should be a class se we can override?


@contextlib.contextmanager
def directory(path: str) -> Iterable[bool]:
    exists_already = os.path.exists(path)
    if not exists_already:
        os.makedirs(path)

    cwd = os.getcwd()
    os.chdir(path)

    try:
        yield exists_already
    finally:
        os.chdir(cwd)


def log_comment(comment: str):
    latest_comment_path = "../latest/comment.txt"
    if os.path.exists("../latest/comment.txt"):
        with open(latest_comment_path, "r") as f:
            assert (
                f.read() != comment
            ), "Comment did not change between runs! Write a new comment or delete the old one."

    with open("comment.txt", "w") as f:
        f.write(comment)


def log_architecture(model: nn.Module):
    with open("architecture.txt", "w") as f:
        f.write(str(model))


def log_pid() -> None:
    with open("pid.txt", "w") as f:
        f.write(str(os.getpid()))


def log_max_memory() -> None:
    with open("max_memory.txt", "w") as f:
        f.write(
            str(torch.cuda.max_memory_allocated())
        )  # todo generalize to multiple devices


def log_last_modified() -> None:
    with open("last_modified.txt", "w") as f:
        f.write(
            datetime.now()
            .isoformat(timespec="seconds")
            .replace(":", "")
            .replace("-", "")
        )


def log_config() -> None:
    shutil.copy(main.__file__, "config.py")


def log_losses(losses: pd.DataFrame) -> None:
    losses.to_csv("losses.csv", index=False)


def log_weights(model: nn.Module) -> None:
    torch.save(model.state_dict(), "weights.pt")


def log_model(model: nn.Module) -> None:
    torch.save(model, "model.pt")


def plot_losses(window_size: int) -> None:
    # todo losses from the .csv file
    losses = pd.read_csv("losses.csv")
    os.makedirs("plots", exist_ok=True)

    images = []

    colors = px.colors.qualitative.Plotly

    for col in losses.columns:
        fig = px.line(
            losses[col],
            title=f"Total weighted objective per iteration (window size {window_size})"
            if col == "objective"
            else f"Average '{col}' loss per iteration (window size {window_size})",
            y=col,
            log_y=True,
            template="plotly_dark",
            color_discrete_sequence=colors[1:] if col == "objective" else colors,
        ).update_layout(showlegend=False, xaxis_title="iteration")
        fig.write_image(f"plots/losses_{col}.png")
        fig.write_html(f"plots/losses_{col}.html")
        images.append(read_image(f"plots/losses_{col}.png"))

    save_image(torch.stack(images) / 255, "losses.png", padding=0, nrow=1)


def plot_scores() -> None:
    scores = pd.read_csv("scores.csv")
    os.makedirs("plots", exist_ok=True)

    images = []

    for col in scores.columns:
        xs = list(range(1, len(scores[col]) + 1))
        fig = (
            px.line(
                scores[col],
                x=xs,
                y=col,
                title=f"'{col}' score per checkpoint",
                log_y=True,
                template="plotly_dark",
            )
            .update_traces(mode="lines+markers")
            .update_layout(showlegend=False, xaxis_title="checkpoint")
            .update_xaxes(tickmode="array", tickvals=xs)
        )

        fig.write_image(f"plots/scores_{col}.png")
        fig.write_html(f"plots/scores_{col}.html")

        images.append(read_image(f"plots/scores_{col}.png"))

    save_image(torch.stack(images) / 255, "scores.png", padding=0, nrow=1)


def log_scores(scores: pd.DataFrame) -> None:
    scores.mean().to_frame().T.to_csv(
        "scores.csv", mode="a", index=False, header=not os.path.exists("scores.csv")
    )


def log_predictions(j: int, preds: torch.Tensor):
    if preds.ndim == 4:
        for k, image in enumerate(preds):
            name = f"{j*preds.shape[0]+k:06d}"
            directory = name[:4]
            os.makedirs(directory, exist_ok=True)
            save_image(image, f"{directory}/{name}.png")


def create_symbolic_link(name: str):
    with directory(".."):
        if os.path.exists("latest"):
            os.unlink("latest")
        os.symlink(name.split("/")[-1], "latest")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['torch_logs']

package_data = \
{'': ['*']}

install_requires = \
['pandas>=1.4.2,<2.0.0',
 'plotly>=5.8.0,<6.0.0',
 'torch>=1.11.0,<2.0.0',
 'torchvision>=0.12.0,<0.13.0']

setup_kwargs = {
    'name': 'torch-logs',
    'version': '0.2.2',
    'description': '',
    'long_description': None,
    'author': 'yohan-pg',
    'author_email': 'pg.yohan@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



from .imports import *

from .low_level import *
from .utils import tee_output


class LogEvent(enum.Enum):
    INIT = "INIT"
    PROGRESS = "PROGRESS"
    CHECKPOINT = "CHECKPOINT"


Evaluation = Callable[[], Iterable[tuple[torch.Tensor, pd.DataFrame]]]


def simple_logging(
    model: nn.Module,
    *,
    name: str,
    evaluation: Optional[Evaluation],
    comment: Optional[str] = None,
    progress_freq: int = 200,
    checkpoint_freq: int = 200,
    heartbeat_freq: int = 20,
    window_size: int = 1,
) -> Callable[[int, pd.DataFrame], Iterable[LogEvent]]:
    def logging_step(
        iteration: int,
        losses: pd.DataFrame,
    ) -> Iterable[LogEvent]:
        with directory(name) as exists_already:
            if not exists_already:
                init_logging(name, model, comment)
                yield LogEvent.INIT

            if iteration % progress_freq == 0:
                yield progress_logging(losses, window_size)

            if iteration % checkpoint_freq == 0:
                with directory(
                    f"checkpoints/{iteration//checkpoint_freq:05d}"
                ):  # todo move inside, tricky
                    yield checkpoint_logging(model, evaluation)

            if iteration % heartbeat_freq == 0:
                heartbeat_logging()

    return logging_step


def init_logging(name: str, model: nn.Module, comment: Optional[str]) -> LogEvent:
    if "pytest" not in sys.argv[0]: # todo figure out why this fails in tests
        tee_output("out.log")

    log_pid()
    log_config()
    log_max_memory()

    log_architecture(model)

    if comment is not None:
        log_comment(comment)

    create_symbolic_link(name)

    return LogEvent.INIT


def progress_logging(losses: pd.DataFrame, window_size: int) -> LogEvent:
    losses = losses.rolling(window_size).mean().iloc[::window_size, :][1:]
    # Some windowing operations also support an online method after constructing a windowing object which returns a new object that supports passing in new DataFrame or Series objects to continue the windowing calculation with the new values (i.e. online calculations).

    log_losses(losses)
    plot_losses(window_size)

    return LogEvent.PROGRESS


def checkpoint_logging(model: nn.Module, evaluation: Evaluation) -> LogEvent:
    if evaluation is not None:
        for j, (preds, scores) in enumerate(evaluation()):
            with directory("predictions"):
                log_predictions(j, preds)

        with directory("../.."):
            log_scores(scores)
            plot_scores()

    log_model(model)
    log_weights(model)

    return LogEvent.CHECKPOINT


def heartbeat_logging():
    log_last_modified()
    log_max_memory()




import os
import contextlib
import torch
import torch.nn as nn
import pandas as pd
import plotly.express as px
import shutil
import enum
from typing import *  # type: ignore
import __main__ as main
import subprocess
import sys
from torchvision.utils import save_image
from torchvision.io import read_image
import datetime